/*A Circular Queue Implementation. (Dynamically expandable)
 *
 * vamsik 12/2/2008
 */
#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<assert.h>
#include "CQueue.h"
/*TODO: Also support defragment. Support CQ_STATS*/
void InitializeCQ(CQueue *cq,unsigned long type_size){
	cq->cq_array = NULL;
	cq->start = 0;
	cq->end = 0;
	cq->capacity = 0;
	cq->type_size = type_size;
	cq->array_size = 0;
#ifdef CQ_STATS
	cq->max_capacity=0;
#endif
}
/*User should create this type and set the type_size*/
void AppendCQ(CQueue *cq,void *input_node){
	void *cptr = NULL;
	void *cptr1 = NULL;
	unsigned long cq_size;

	if(cq->capacity == cq->array_size){  /*Initial Circular Queue or Full Queue*/

		cptr = (char *) malloc((cq->type_size)*
			(cq->array_size+CQUEUE_INCREMENT_SIZE));
		assert(cptr);

		if(cq->cq_array){ /*OVERFLOW: Readjust the Circular Queue*/
			/*------(start)<----->*/
			cq_size = cq->array_size - cq->start;
			cptr1 = ((char *)cq->cq_array+((cq->start)*
				(cq->type_size)));
			assert(memcpy(cptr,cptr1,(cq_size)*
				(cq->type_size)) == cptr);
			/*<----->(start)------*/
			cptr1 = (char *)cptr + (cq_size)*(cq->type_size);
			assert(memcpy(cptr1,cq->cq_array,
				(cq->start+1)*(cq->type_size)) == cptr1);
			/*update the start and end*/
			cq->start = 0;
			cq->end = cq->array_size;
			free(cq->cq_array); 
			cq->cq_array = cptr;
		}else{
			cq->cq_array = cptr;
			cq->start = 0; cq->end = 0;
		}
		cq->array_size += CQUEUE_INCREMENT_SIZE;
	}

	/*Append*/
	cptr1 = (char *)cq->cq_array + ((cq->end)*(cq->type_size));
	assert(memcpy(cptr1,input_node,cq->type_size) == cptr1);
	/*Increment the end*/
	cq->end = (cq->end+1)%(cq->array_size);
	cq->capacity++;

}

/*NOTE: If dequeue is able to copy returns 1 else returns 0 */
char DequeueCQ(CQueue *cq, void *copy_to){
	void *cptr = NULL;
	if(!cq->capacity){
		return 0; 
	}
	cptr = (char *)cq->cq_array+((cq->start)*(cq->type_size));
	if((memcpy(copy_to,cptr,cq->type_size) == copy_to)){
		cq->start = (cq->start+1)%(cq->array_size);
		(cq->capacity)--;
		return 1;
	}
	return 0;
}

void DeleteCQ(CQueue *cq){
	free(cq->cq_array);
	cq->capacity = 0;
	cq->start = 0; cq->end=0;
}

#ifndef NO_UNIT_TEST_CQ
int main(int argc,char **argv){
	CQueue cq;
	double d = -1.5e+15;
	double epsilon = -0.5e+15;
	unsigned int i;
	double d_copy;
	InitializeCQ(&cq,sizeof(double));
	for(i=0;i<(1<<20);i++){
		d+=epsilon;
		AppendCQ(&cq,&d);
		printf("[ENQUEUE] %E \n",d);
		if(!(i%50) && DequeueCQ(&cq,&d_copy)){
			printf("*[DEQUEUE] %E \n",d_copy);
		}
	}
	printf("The capacity of CQ is %lu \n",cq.capacity);
	for(i=0;i<(1<<20);i++){
		if(DequeueCQ(&cq,&d_copy)){
			printf("%E \n",d_copy);
		}else{
			break;
		}
	}
	d=18.0;
	AppendCQ(&cq,&d);
	DequeueCQ(&cq,&d_copy);
	printf("%E \n",d_copy);
	printf("%s \n",DequeueCQ(&cq,&d_copy));

}

#endif
